from __future__ import annotations

import os
from typing import List, Optional, Union

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
from pydantic import BaseModel

from apparun.impact_model import ImpactModel

RESULTS = {}


def register_result(result_name):
    """
    This decorator registers a new ImpactModelResult class in RESULTS registry.
    :param result_name: name of the new result to register
    :return: new ImpactModelResult class
    """

    def decorator(decorated_class):
        if result_name not in RESULTS:
            RESULTS[result_name] = decorated_class
        return decorated_class

    return decorator


def get_result(result_name: str):
    """
    Get a registered ImpactModelResult class by name.
    :param result_name: registered name of the desired ImpactModelResult.
    :return: registered ImpactModelResult class corresponding to the name.
    """
    return RESULTS[result_name]


def registered_results() -> List[str]:
    """
    Get a list of registered ImpactModelResult names.
    :return: list of registered ImpactModelResult names.
    """
    return list(RESULTS.keys())


class ImpactModelResult(BaseModel):
    """
    An impact model result is one, or a collection of tables and/or figures generated
    by executing an impact model.
    """

    output_name: str
    impact_model: ImpactModel
    html_save_path: Optional[str] = None
    pdf_save_path: Optional[str] = None
    png_save_path: Optional[str] = None
    table_save_path: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None

    def run(self):
        """
        Execute the result to generate wanted outputs.
        """
        table = self.get_table()
        self.get_figure(table)

    def get_table(self) -> Union[pd.DataFrame, pd.Series]:
        """
        Abstract method.
        Generate the output as a table, or a collection of tables.
        :return: tabular results as a pandas DataFrame.
        """
        return pd.DataFrame()

    def get_figure(self, table: pd.DataFrame):
        """
        Abstract method.
        Generate the output as a figures, or a collection of figures, using the output
        of the get_table method.
        :param table: tabular result data.
        :return: figure, or collection of figure generated.
        """
        return None

    def save_figure(self, fig, name_suffix=None):
        """
        Save a figure to disk, according to the configuration specified in result
        attributes.
        :param fig: figure to save
        :param name_suffix: optional file name suffix.
        """
        filename = (
            self.output_name + "-" + name_suffix
            if name_suffix is not None
            else self.output_name
        )
        if self.html_save_path is not None:
            os.makedirs(self.html_save_path, exist_ok=True)
            figure_path = os.path.join(self.html_save_path, f"{filename}.html")
            fig.write_html(figure_path)
        if self.pdf_save_path is not None:
            os.makedirs(self.pdf_save_path, exist_ok=True)
            figure_path = os.path.join(self.pdf_save_path, f"{filename}.pdf")
            fig.write_image(figure_path, width=self.width, height=self.height)
        if self.png_save_path is not None:
            os.makedirs(self.png_save_path, exist_ok=True)
            figure_path = os.path.join(self.png_save_path, f"{filename}.png")
            fig.write_image(figure_path, width=self.width, height=self.height)


@register_result("tree_map")
class TreeMapResult(ImpactModelResult):
    """
    Generate a TreeMap for each impact, representing the contribution of all nodes to
    the root node result.
    See https://plotly.com/python/treemaps/ for more information.
    """

    parameters: Optional[dict[str, Union[float, str]]] = {}

    def get_table(self) -> pd.DataFrame:
        """
        Generate treemap output as a table, or a collection of tables.
        Save it to disk according to configuration specified in result attributes.
        :return: tabular results as a pandas DataFrame.
        """
        node_scores = self.impact_model.get_nodes_scores(**self.parameters)
        node_scores = [node_score.to_unpivoted_df() for node_score in node_scores]
        node_scores = pd.concat(node_scores)
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            node_scores.to_csv(figure_path)
        return node_scores

    def get_figure(self, table: pd.DataFrame):
        """
        Generate one distinct treemap per impact method.
        Save figure(s) to disk, according to the configuration specified in result
        attributes.
        :param table: tabular treemap result data.
        :return: figure, or collection of figure generated.
        """
        figs = []
        for method_name in pd.unique(table["method"]):
            method_data = table[table["method"] == method_name]
            fig = go.Figure(
                go.Treemap(
                    labels=method_data["name"].values,
                    parents=method_data["parent"].values,
                    values=method_data["score"].values,
                    branchvalues="total",
                    root_color="lightgrey",
                )
            )
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method_name)
            figs.append(fig)
        return figs


@register_result("sankey")
class SankeyDiagramResult(ImpactModelResult):
    """
    Generate a Sankey diagram for each impact, representing the contribution of all
    nodes to the root node result.
    See https://plotly.com/python/sankey-diagram/ for more information.
    """

    parameters: Optional[dict[str, Union[float, str]]] = {}

    def get_table(self) -> pd.DataFrame:
        """
        Generate Sankey output as a table, or a collection of tables.
        Save it to disk according to configuration specified in result attributes.
        :return: tabular results as a pandas DataFrame.
        """
        node_scores = self.impact_model.get_nodes_scores(**self.parameters)
        node_scores = [node_score.to_unpivoted_df() for node_score in node_scores]
        node_scores = pd.concat(node_scores)
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            node_scores.to_csv(figure_path)
        names_to_indices = {
            node_scores["name"].unique()[i]: i
            for i in range(node_scores["name"].unique().shape[0])
        }
        node_scores["source"] = node_scores["name"]
        node_scores = node_scores.replace(
            {"source": names_to_indices, "parent": names_to_indices}
        )
        return node_scores

    def get_figure(self, table: pd.DataFrame) -> List[go.Figure]:
        """
        Generate one distinct Sankey diagram per impact method.
        Save figure(s) to disk, according to the configuration specified in result
        attributes.
        :param table: tabular sankey result data.
        :return: figure, or collection of figure generated.
        """
        figs = []
        for method_name in pd.unique(table["method"]):
            method_data = table[table["method"] == method_name]
            fig = go.Figure(
                data=[
                    go.Sankey(
                        node=dict(
                            pad=15,
                            thickness=20,
                            line=dict(color="black", width=0.5),
                            label=method_data["name"].values,
                            color="blue",
                        ),
                        link=dict(
                            source=method_data["source"].values,
                            target=method_data["parent"].values,
                            value=method_data["score"].values,
                        ),
                    )
                ]
            )
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method_name)
            figs.append(fig)
        return figs


@register_result("sobol")
class SobolIndexResult(ImpactModelResult):
    """
    Generate sobol S1 index for each impact as a heatmap. Sobol S1 indices represent
    the first order contribution of each model's parameter to the score variance.
    See https://plotly.com/python/sankey-diagram/ for more information.
    """

    parameters: Optional[dict[str, Union[float, str]]] = None
    n: int

    def get_table(self) -> pd.DataFrame:
        """
        Generate sobol S1 indices.
        Save it to disk according to configuration specified in result attributes.
        :return: tabular results as a pandas DataFrame.
        """
        sobol_indices = self.impact_model.get_sobol_s1_indices(n=self.n)
        table = pd.DataFrame(sobol_indices)
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            table.to_csv(figure_path)
        return table

    def get_figure(self, table: pd.DataFrame):
        """
        Generate a heatmap for S1 sobol indices of each model's parameter, and for each
        impact method.
        Save figure(s) to disk, according to the configuration specified in result
        attributes.
        :param table: tabular sankey result data.
        :return: figure, or collection of figure generated.
        """
        pivoted_table = table.pivot(
            index="parameter", columns="method", values="sobol_s1"
        )
        fig = px.imshow(pivoted_table, text_auto=True)
        pio.full_figure_for_development(fig, warn=False)
        self.save_figure(fig)
        return fig


@register_result("nodes_sobol")
class NodesSobolIndexResult(ImpactModelResult):
    """ """

    parameters: Optional[dict[str, Union[float, str]]] = None
    n: int

    def get_table(self) -> pd.DataFrame:
        """ """
        sobol_indices = self.impact_model.get_sobol_s1_indices(n=self.n, all_nodes=True)
        table = pd.DataFrame(sobol_indices)
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            table.to_csv(figure_path)
        return table

    def get_figure(self, table: pd.DataFrame):
        """ """
        figures = []
        for method in pd.unique(table["method"]):
            pivoted_table = table[table["method"] == method]
            pivoted_table = pivoted_table.pivot(
                index="parameter", columns="node", values="sobol_s1"
            )
            fig = px.imshow(pivoted_table, text_auto=True)
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method)
            figures.append(fig)
        return figures


@register_result("nodes_uncertainty")
class NodesUncertaintyResult(ImpactModelResult):
    """
    Generate uncertainty for each node using Monte Carlo. Result figure as a boxplot.
    """

    n: int

    def get_table(self) -> pd.DataFrame:
        """
        Run monte carlo simulation for each node, get all values as a long format table.
        :return: results of each draw for each node as a long format table
        """
        nodes_scores = self.impact_model.get_uncertainty_nodes_scores(n=self.n)
        nodes_scores = [node_scores.to_unpivoted_df() for node_scores in nodes_scores]
        table = pd.concat(nodes_scores)
        table = table.rename(columns={"name": "node"})
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            table.to_csv(figure_path)
        return table

    def get_figure(self, table: pd.DataFrame):
        """
        Display uncertainty result of each node with boxplots, one figure per impact.
        :param table: results of each draw for each node as a long format table
        :return: all figures generated
        """
        figures = []
        for method in pd.unique(table["method"]):
            fig = px.box(table[table["method"] == method], x="node", y="score")
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method)
            figures.append(fig)
        return figures


@register_result("uncertainty")
class UncertaintyResult(ImpactModelResult):
    n: int
    """
    Generate uncertainty for FU using Monte Carlo. Result figure as a boxplot.
    """

    def get_table(self) -> pd.DataFrame:
        """
        Run monte carlo simulation for FU, get all values as a long format table.
        :return: results of each draw as a long format table
        """
        lcia_score = self.impact_model.get_uncertainty_scores(n=self.n)
        lcia_score = lcia_score.to_unpivoted_df()
        lcia_score = lcia_score.rename(columns={"name": "node"})
        lcia_score["node"] = "fu"
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            lcia_score.to_csv(figure_path)
        return lcia_score

    def get_figure(self, table: pd.DataFrame):
        """
        Display uncertainty result of FU with boxplot, one figure per impact.
        :param table: results of each draw for each node as a long format table
        :return: all figures generated
        """
        figures = []
        for method in pd.unique(table["method"]):
            fig = px.box(table[table["method"] == method], x="node", y="score")
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method)
            figures.append(fig)
        return figures



from __future__ import annotations

from typing import Dict, List, Optional, Set, Union

import pandas as pd
from pydantic import BaseModel


class LCIAScores(BaseModel):
    """
    Scores for each impact method.
    """

    scores: Optional[Dict[str, Union[float, List[float]]]] = {}

    @property
    def method_names(self) -> Set[str]:
        """
        Get all LCIA methods assessed.
        :return: LCIA methods assessed
        """
        return set(self.scores.keys())

    def to_unpivoted_df(self) -> pd.DataFrame:
        if isinstance(list(self.scores.values())[0], float) or isinstance(
            list(self.scores.values())[0], float
        ):
            df = pd.DataFrame(self.scores, index=[0])
        else:
            df = pd.DataFrame(self.scores)
        df = pd.melt(df, var_name="method", value_name="score")
        return df

    @staticmethod
    def sum(lcia_scores: List[LCIAScores]) -> LCIAScores:
        """
        Sum element-wise all scores for each method.
        :param lcia_scores: LCIA scores to sum up.
        :return: summed LCIA scores
        """
        scores = {
            method_name: [lcia_score.scores[method_name] for lcia_score in lcia_scores]
            for method_name in lcia_scores[0].method_names
        }
        scores = {
            method_name: sum(score)
            if isinstance(score[0], float)
            else [sum(x) for x in zip(*score)]
            for method_name, score in scores.items()
        }
        return LCIAScores(scores=scores)



from __future__ import annotations

from typing import Callable, Dict, List, Optional, Union

import numpy as np
import yaml
from pydantic import BaseModel
from SALib.analyze import sobol

from apparun.impact_tree import ImpactTreeNode
from apparun.parameters import ImpactModelParams
from apparun.score import LCIAScores
from apparun.tree_node import NodeScores


class LcaPractitioner(BaseModel):
    """
    Information about a LCA practitioner.
    """

    name: Optional[str] = None
    organization: Optional[str] = None
    mail: Optional[str] = None


class LcaStudy(BaseModel):
    """
    Information about LCA study, in order to understand its scope and for
    reproducibility.
    """

    link: Optional[str] = None
    description: Optional[str] = None
    date: Optional[str] = None
    version: Optional[str] = None
    license: Optional[str] = None
    appabuild_version: Optional[str] = None


class ModelMetadata(BaseModel):
    """
    Contain information various information about the context of production of the
    impact model.
    """

    author: Optional[LcaPractitioner] = None
    reviewer: Optional[LcaPractitioner] = None
    report: Optional[LcaStudy] = None

    def to_dict(self):
        """
        Convert self to dict.
        :return: self as a dict
        """
        return self.model_dump()

    @staticmethod
    def from_dict(model_metadata: dict) -> ModelMetadata:
        """
        Convert dict to ModelMetadata object.
        :param model_metadata: dict containing construction parameters of the model
        metadata.
        :return: constructed model metadata.
        """
        return ModelMetadata(
            author=LcaPractitioner(**model_metadata["author"]),
            reviewer=LcaPractitioner(**model_metadata["reviewer"]),
            report=LcaStudy(**model_metadata["report"]),
        )


class ImpactModel(BaseModel):
    """
    Impact model contains all the information required to compute the impact of an
    LCA built with Appa Build.
    """

    metadata: Optional[ModelMetadata] = None
    parameters: Optional[ImpactModelParams] = None
    tree: Optional[ImpactTreeNode] = None

    @property
    def name(self):
        return self.tree.name

    @property
    def transformation_table(
        self,
    ) -> Dict[str, Callable[[Union[str, float]], Dict[str, float]]]:
        """
        Map each parameter to its transform method.
        :return: a dict mapping impact model's parameters' name with their transform
        method.
        """
        return {parameter.name: parameter.transform for parameter in self.parameters}

    def transform_parameters(
        self, parameters: Dict[str, Union[List[Union[str, float]], Union[str, float]]]
    ) -> Dict[str, Union[List[Union[str, float]], Union[str, float]]]:
        """
        Transform all the parameters' values, so it can be fed into a node's compute
        method. See ImpactModelParam's transform methods for more information.
        :param parameters: a dict mapping parameters' name and parameters' value, or
        list of values.
        :return: a dict mapping parameters' name and parameters' transformed value, or
        list of transformed values.
        """
        list_parameters = {
            name: parameter
            for name, parameter in parameters.items()
            if isinstance(parameter, list)
        }
        single_parameters = {
            name: parameter
            for name, parameter in parameters.items()
            if not isinstance(parameter, list)
        }
        if len(list_parameters) == 0:
            return {
                name: value
                for table in [
                    self.transformation_table[parameter_name](parameter_value)
                    for parameter_name, parameter_value in parameters.items()
                ]
                for name, value in table.items()
            }
        assert min(
            len(list_parameter) for list_parameter in list_parameters.values()
        ) == max(len(list_parameter) for list_parameter in list_parameters.values())
        full_list_parameters = {
            **{
                parameter_name: [parameter_value]
                * len(list(list_parameters.values())[0])
                for parameter_name, parameter_value in single_parameters.items()
            },
            **list_parameters,
        }
        return {
            name: value
            for table in [
                self.transformation_table[parameter_name](parameter_value)
                for parameter_name, parameter_value in full_list_parameters.items()
            ]
            for name, value in table.items()
        }

    def to_dict(self):
        """
        Convert self to dict.
        :return: self as a dict
        """
        return {
            "metadata": self.metadata.to_dict(),
            "parameters": self.parameters.to_list(sorted_by_name=True),
            "tree": self.tree.to_dict(),
        }

    def to_yaml(self, filepath: str, compile_models: bool = True):
        """
        Convert self to yaml file.
        :param filepath: filepath of the yaml file to create.
        :param compile_models: if True, all models in tree nodes will be compiled.
        ImpactModel will be bigger, but its execution will be faster at first use.
        """
        if compile_models:
            self.tree.compile_models()
        with open(filepath, "w") as stream:
            yaml.dump(self.to_dict(), stream, sort_keys=False)

    @staticmethod
    def from_dict(impact_model: dict) -> ImpactModel:
        """
        Convert dict to ImpactModel object.
        :param impact_model: dict containing construction parameters of the impact
        model.
        :return: constructed impact model.
        """
        return ImpactModel(
            metadata=ModelMetadata.from_dict(impact_model["metadata"]),
            parameters=ImpactModelParams.from_list(impact_model["parameters"]),
            tree=ImpactTreeNode.from_dict(impact_model["tree"]),
        )

    def from_tree_children(self) -> List[ImpactModel]:
        """
        Create a new impact model for each of Impact Model tree root node's children.
        Parameters of the impact model are copied, so unused parameters can remain in
        newly created impact models.
        :return: a list of newly created impact models.
        """
        return [
            ImpactModel(parameters=self.parameters, tree=child)
            for child in self.tree.children
        ]

    @staticmethod
    def from_yaml(filepath: str) -> ImpactModel:
        """
        Convert a yaml file to an ImpactModel object.
        :param filepath: yaml file containing construction parameters of the impact
        model.
        :return: constructed impact model.
        """
        with open(filepath, "r") as stream:
            impact_model = yaml.safe_load(stream)
            return ImpactModel.from_dict(impact_model)

    def get_scores(self, **params) -> LCIAScores:
        """
        Get impact scores of the root node for each impact method, according to the
        parameters.
        :param params: value, or list of values of the impact model's parameters.
        List of values must have the same length. If single values are provided
        alongside a list of values, it will be duplicated to the appropriate length.
        :return: a dict mapping impact names and corresponding score, or list of scores.
        """
        missing_params = self.parameters.get_missing_parameter_names(params)
        default_params = self.parameters.get_default_values(missing_params)
        transformed_params = self.transform_parameters({**params, **default_params})
        scores = self.tree.compute(transformed_params)
        return scores

    def get_nodes_scores(
        self, by_property: Optional[str] = None, **params
    ) -> List[NodeScores]:
        """
        Get impact scores of the each node for each impact method, according to the
        parameters.
        :param by_property: if different than None, results will be pooled by nodes
        sharing the same property value. Property name is the value of by_property.
        :param params: value, or list of values of the impact model's parameters.
        List of values must have the same length. If single values are provided
        alongside a list of values, it will be duplicated to the appropriate length.
        :return: a list of dict mapping impact names and corresponding score, or list
        of scores, for each node/property value.
        """
        missing_params = self.parameters.get_missing_parameter_names(params)
        default_params = self.parameters.get_default_values(missing_params)
        transformed_params = self.transform_parameters({**params, **default_params})
        scores = [
            NodeScores(
                name=node.name,
                properties=node.properties,
                parent=node.parent.name if node.parent is not None else "",
                lcia_scores=node.compute(
                    transformed_params, direct_impacts=by_property is not None
                ),
            )
            for node in self.tree.unnested_descendants
        ]
        if by_property is not None:
            scores = NodeScores.combine_by_property(scores, by_property)
        return scores

    def get_uncertainty_nodes_scores(self, n) -> List[NodeScores]:
        """ """
        samples = self.parameters.uniform_draw(n)
        samples = self.parameters.draw_to_distrib(samples)
        nodes_scores = self.get_nodes_scores(**samples)
        return nodes_scores

    def get_uncertainty_scores(self, n) -> LCIAScores:
        """ """
        samples = self.parameters.uniform_draw(n)
        samples = self.parameters.draw_to_distrib(samples)
        lcia_scores = self.get_scores(**samples)
        return lcia_scores

    def get_sobol_s1_indices(
        self, n, all_nodes: bool = False
    ) -> List[Dict[str, Union[str, np.ndarray]]]:
        """
        Get sobol first indices, which corresponds to the contribution of each
        parameter to total result variance.
        :param n: number of samples to draw with monte carlo.
        :param all_nodes: if True, sobol s1 indices will be computed for each node. Else,
        only for root node (FU).
        :return: unpivoted dataframe containing sobol first indices for each parameter,
        impact method, and node name if all_nodes is True.
        """
        samples = self.parameters.sobol_draw(n)
        samples = self.parameters.draw_to_distrib(samples)
        if all_nodes:
            lcia_scores = self.get_nodes_scores(**samples)
            sobol_s1_indices = []
            for node_scores in lcia_scores:
                for method, scores in node_scores.lcia_scores.scores.items():
                    s1 = sobol.analyze(
                        self.parameters.sobol_problem,
                        np.array(scores),
                        calc_second_order=True,
                    )["S1"]
                    sobol_s1_indices += [
                        {
                            "node": node_scores.name,
                            "method": method,
                            "parameter": self.parameters.sobol_problem["names"][i],
                            "sobol_s1": s1[i],
                        }
                        for i in range(len(s1))
                    ]
            return sobol_s1_indices
        lcia_scores = self.get_scores(**samples)
        sobol_s1_indices = []
        for method, scores in lcia_scores.scores.items():
            s1 = sobol.analyze(
                self.parameters.sobol_problem, np.array(scores), calc_second_order=True
            )["S1"]
            sobol_s1_indices += [
                {
                    "node": self.tree.name,
                    "method": method,
                    "parameter": self.parameters.sobol_problem["names"][i],
                    "sobol_s1": s1[i],
                }
                for i in range(len(s1))
            ]
        return sobol_s1_indices



import os

import pandas as pd
import plotly.express as px
import plotly.io as pio

from apparun.impact_model import ImpactModel
from apparun.results import (
    ImpactModelResult,
    NodesUncertaintyResult,
    get_result,
    register_result,
)

OUTPUT_FILES_PATH = "outputs/"

# This script aims to illustrate how to use main functions of Appa Run. We will
# compute impacts, generate figures and tables, and create user-defined custom results.

impact_model = ImpactModel.from_yaml("samples/impact_models/nvidia_ai_gpu_chip.yaml")

# Get scores and scores for each node
scores = impact_model.get_scores(lifespan=3,
                                 architecture="Maxwell",
                                 cuda_core=[256, 512, 1024],
                                 energy_per_inference=[0.05, 0.06, 0.065])
print(scores)
nodes_scores = impact_model.get_nodes_scores(lifespan=3,
                                             architecture="Maxwell",
                                             cuda_core=[256, 512, 1024],
                                             energy_per_inference=[0.05, 0.06, 0.065])
print(nodes_scores)

# Generate Tree Map figures for nodes
# We can update default values
# of impact models are changed beforehand. New min and max will be automatically adapted
# using pm and pm_perc.
impact_model.parameters.update_defaults({"cuda_core": 256, "architecture": "Maxwell"})

tree_map_result = get_result("tree_map")(
    impact_model=impact_model,
    output_name="tree_map",
    pdf_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
    table_save_path=os.path.join(OUTPUT_FILES_PATH, "tables/"),
    html_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
)
tree_map_table = tree_map_result.get_table()
tree_map_result.get_figure(tree_map_table)

# Generate Sankey figures for nodes
sankey_result = get_result("sankey")(
    impact_model=impact_model,
    output_name="sankey",
    pdf_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
    table_save_path=os.path.join(OUTPUT_FILES_PATH, "tables/"),
    html_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
)
sankey_table = sankey_result.get_table()
sankey_result.get_figure(sankey_table)

# Generate Sobol figures for nodes
sobol_result = get_result("sobol")(
    impact_model=impact_model,
    n=4096,
    output_name="sobol",
    pdf_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
    table_save_path=os.path.join(OUTPUT_FILES_PATH, "tables/"),
    html_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
)
sobol_table = sobol_result.get_table()
sobol_result.get_figure(sobol_table)

# Generate uncertainty figure for nodes
uncertainty_nodes_result = get_result("nodes_uncertainty")(
    impact_model=impact_model,
    n=4096,
    output_name="uncertainty_nodes",
    pdf_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
    table_save_path=os.path.join(OUTPUT_FILES_PATH, "tables/"),
    html_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
)
uncertainty_nodes_table = uncertainty_nodes_result.get_table()
uncertainty_nodes_result.get_figure(uncertainty_nodes_table)


# New types of results can be generated in user script, without modifying Appa Run
# source code, thanks to register_result decorator.


@register_result("new_nodes_uncertainty")
class NewNodesUncertaintyResult(NodesUncertaintyResult):
    """
    Generate uncertainty for each node using Monte Carlo. Result figure as a boxplot.
    """

    def get_figure(self, table: pd.DataFrame):
        """ """
        figures = []
        for method in pd.unique(table["method"]):
            fig = px.box(table[table["method"] == method], x="score", y="node")
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method)
            figures.append(fig)
        return figures


# Let's compute nodes uncertainty again, with custom class.

@register_result("uncertainty")
class UncertaintyResult(ImpactModelResult):
    n: int
    """
    Generate uncertainty for each node using Monte Carlo. Result figure as a boxplot.
    """

    def get_table(self) -> pd.DataFrame:
        """ """
        lcia_score = self.impact_model.get_uncertainty_scores(n=self.n)
        lcia_score = lcia_score.to_unpivoted_df()
        lcia_score = lcia_score.rename(columns={"name": "node"})
        lcia_score["node"] = "fu"
        if self.table_save_path is not None:
            os.makedirs(self.table_save_path, exist_ok=True)
            figure_path = os.path.join(self.table_save_path, f"{self.output_name}.csv")
            lcia_score.to_csv(figure_path)
        return lcia_score

    def get_figure(self, table: pd.DataFrame):
        """ """
        figures = []
        for method in pd.unique(table["method"]):
            fig = px.box(table[table["method"] == method], x="node", y="score")
            pio.full_figure_for_development(fig, warn=False)
            self.save_figure(fig, name_suffix=method)
            figures.append(fig)
        return figures


new_nodes_uncertainty_result = get_result("new_nodes_uncertainty")(
    impact_model=impact_model,
    n=4096,
    output_name="new_uncertainty_nodes",
    pdf_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
    table_save_path=os.path.join(OUTPUT_FILES_PATH, "tables/"),
    html_save_path=os.path.join(OUTPUT_FILES_PATH, "figures/"),
)
new_nodes_uncertainty_table = new_nodes_uncertainty_result.get_table()
new_nodes_uncertainty_result.get_figure(new_nodes_uncertainty_table)



import logging
import os
import time
from functools import wraps
from typing import Dict, List, Union

from apparun import results
from apparun.impact_model import ImpactModel
from apparun.results import get_result

IMPACTS_MODEL_DIR = os.environ.get("APPARUN_IMPACT_MODELS_DIR")
logging.basicConfig(
    filename="apparun.log",
    filemode="w",
    format="%(name)s - %(levelname)s - %(message)s",
    level=logging.INFO,
)


def execution_time_logging(func):
    """
    Decorator to log execution time as an info.
    :param func: function to wrap.
    :return: wrapped function.
    """

    @wraps(func)
    def execution_time_logging_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        logging.info(f"Function {func.__name__} executed in {total_time:.4f} seconds")
        return result

    return execution_time_logging_wrapper


@execution_time_logging
def compute_impacts(
    impact_model_name: str, params: Dict, all_nodes: bool = False
) -> Union[List, Dict[str, Union[float, List[float]]]]:
    """
    Load an impact model from disk from its name, and get impact scores of the root node
    for each impact method, according to the parameters.
    APPARUN_IMPACT_MODELS_DIR environment variable should be specified (see README.md).
    :param impact_model_name: name of the impact model to load
    :param params: value, or list of values of the impact model's parameters.
    List of values must have the same length. If single values are provided
    alongside a list of values, it will be duplicated to the appropriate length.
    :param all_nodes: if True, scores will be computed for each node. Only root node
    otherwise (default).
    :return: a dict mapping impact names and corresponding score, or list of scores.
    """
    impact_model = ImpactModel.from_yaml(
        os.path.join(IMPACTS_MODEL_DIR, f"{impact_model_name}.yaml")
    )
    if all_nodes:
        return impact_model.get_nodes_scores(**params)
    return dict(impact_model.get_scores(**params))


@execution_time_logging
def get_valid_models() -> List[str]:
    """
    Get a list of all valid impact models in the directory specified by
    APPARUN_IMPACT_MODELS_DIR environment variable.
    :return: a list of all valid impact models.
    """
    valid_impact_models = []
    for file in os.listdir(IMPACTS_MODEL_DIR):
        if file.endswith(".yaml"):
            try:
                ImpactModel.from_yaml(os.path.join(IMPACTS_MODEL_DIR, file))
                valid_impact_models.append(file.replace(".yaml", ""))
            except KeyError:
                print(f"{file.replace('.yaml', '')} is not a valid impact model.")
    return valid_impact_models


@execution_time_logging
def get_model_params(impact_model_name: str) -> List[Dict]:
    """
    Return the list of parameters required to execute an impact model.
    APPARUN_IMPACT_MODELS_DIR environment variable should be specified (see README.md).
    :param impact_model_name: name of the impact model to load.
    :return: a list of parameters required by the model.
    """
    impact_model = ImpactModel.from_yaml(
        os.path.join(IMPACTS_MODEL_DIR, f"{impact_model_name}.yaml")
    )
    return [parameter.to_dict() for parameter in impact_model.parameters]


@execution_time_logging
def compute_results(results_config: List[Dict]):
    """
    Generate results according to results_config.
    :param results_config: list of results wanted. Each element of this list will be
    used to construct an ImpactModelResult, using the elements in "args" argument.
    Result subclass is determined by "class" argument.
    """
    for result_config in results_config:
        result_name = result_config["result_name"]
        result_constructor_args = result_config["args"]
        if "parameters" in result_constructor_args["impact_model"]:
            result_constructor_args["parameters"] = result_constructor_args[
                "impact_model"
            ]["parameters"]
        result_constructor_args["impact_model"] = ImpactModel.from_yaml(
            os.path.join(
                IMPACTS_MODEL_DIR,
                f"{result_constructor_args['impact_model']['name']}.yaml",
            )
        )
        result_class = get_result(result_name)
        result_objet = result_class(**result_constructor_args)
        result_objet.run()



from __future__ import annotations

from typing import Dict, List, Optional, Set, Union

import pandas as pd
from pydantic import BaseModel

from apparun.score import LCIAScores


class NodeProperties(BaseModel):
    """
    Properties of an impact model node. Can be used by to break the results down
    according to life cycle phase, for exemple. Properties can be key/value
    (ex: {"phase": "production"} or flags (ex: {production_phase: True}).
    """

    properties: Dict[str, Union[str, float, bool]] = {}

    @classmethod
    def from_dict(
        cls, properties: Dict[str, Union[str, float, bool]]
    ) -> NodeProperties:
        """
        Construct using dict.
        :param properties: properties of the node.
        :return: constructed NodeProperties
        """
        return NodeProperties(properties=properties)

    def get_property_value(self, property_name: str) -> Optional[Union[str, bool]]:
        """
        Gives the value of a property by name, returns None if property doesn't exist.
        :param property_name: name of the property to get value
        :return: value of the property, None if doesn't exist
        """
        return (
            self.properties[property_name] if property_name in self.properties else None
        )


class NodeScores(BaseModel):
    """
    Gathers all useful information to exploit computed node wise LCIA results.
    """

    name: str
    "Node name/property value, if results have been combined by property value."
    parent: str
    "Name of parent node."
    properties: NodeProperties
    "Properties of the node."
    lcia_scores: LCIAScores
    "Computed LCIA scores, for each method."

    @staticmethod
    def combine_by_property(
        nodes_scores: List[NodeScores], property_name: str
    ) -> List[NodeScores]:
        """
        Sum up (element wise) the scores of each node sharing the same property value,
        for each method.
        :param nodes_scores: node scores to combine according to property value.
        :param property_name: name of the property under consideration
        :return: list of newly created nodes. Name of the node will be the property
        value.
        """
        all_values = set(
            [
                node.properties.get_property_value(property_name)
                for node in nodes_scores
                if len(node.properties.properties) > 0
            ]
        )
        nodes_by_value = {
            value: [
                node
                for node in nodes_scores
                if node.properties.get_property_value(property_name) == value
            ]
            for value in all_values
        }
        scores_by_values = {
            value: LCIAScores.sum([node.lcia_scores for node in nodes])
            for value, nodes in nodes_by_value.items()
        }
        return [
            NodeScores(
                name=str(value),
                parent="",
                properties=NodeProperties(),
                lcia_scores=scores_by_values[value],
            )
            for value in all_values
        ]

    def to_unpivoted_df(self) -> pd.DataFrame:
        df = self.lcia_scores.to_unpivoted_df()
        df["name"] = self.name
        df["parent"] = self.parent
        return df



from __future__ import annotations

import itertools
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, List, Optional, Union

import numpy as np
from pydantic import BaseModel
from sympy import Expr, lambdify, parse_expr

from apparun.score import LCIAScores
from apparun.tree_node import NodeProperties


class ImpactTreeNode(BaseModel):
    """
    Impact Model tree node representing the impacts of an activity as well as its
    children.
    """

    name: str
    amount: Optional[Union[Expr, float]] = None
    direct_impacts: Optional[Dict[str, Expr]] = {}
    scaled_direct_impacts: Optional[Dict[str, Expr]] = {}
    models: Optional[Dict[str, Expr]] = {}
    parent: Optional[ImpactTreeNode] = None
    children: Optional[List[ImpactTreeNode]] = []
    properties: NodeProperties = NodeProperties(properties={})
    _raw_direct_impact: Optional[Expr] = None
    _combined_amount: Optional[Union[Expr, float]] = None

    class Config:
        arbitrary_types_allowed = True

    @property
    def models_compiled(self):
        return len(self.models) > 0 and len(self.scaled_direct_impacts) > 0

    @property
    def unnested_descendants(self) -> List[ImpactTreeNode]:
        """
        Walk recursively through node's children to return a list of all its
        descendants.
        :return: a list containing current node and all its descendants.
        """
        return list(
            itertools.chain.from_iterable(
                [child.unnested_descendants for child in self.children]
            )
        ) + [self]

    @property
    def combined_amount(self) -> Union[float, Expr]:
        if self.parent is None:
            return self.amount
        if self._combined_amount is None:
            self._combined_amount = self.amount * self.parent.combined_amount
        return self._combined_amount

    def compile_models(self):
        """
        Replace descendants symbols in node's model by their corresponding model.
        Result is stored in "compiled_model" node's attribute.
        """
        self.scaled_direct_impacts = {
            method: self.combined_amount * direct_impact
            for method, direct_impact in self.direct_impacts.items()
        }
        self.models = self.scaled_direct_impacts
        for child in self.children:
            child.compile_models()
            self.models = {
                method: model + child.models[method]
                for method, model in self.models.items()
            }

    def new_child(self, **args) -> ImpactTreeNode:
        """
        Build a new node as a child.
        :param args: construction parameters of new node.
        :return: constructed node
        """
        child = ImpactTreeNode(**args, parent=self)
        self.children.append(child)
        return child

    def new_child_from_dict(self, child: dict) -> ImpactTreeNode:
        """
        Build a new node as a child.
        :param child: dict containing construction parameters of new node.
        :return: constructed node
        """
        child = ImpactTreeNode.from_dict(child)
        child.parent = self
        self.children.append(child)
        return child

    def name_already_in_tree(self, name: str) -> bool:
        """
        Check if a node exists in descendants by comparing names.
        :param name: name to check presence in descendants
        :return: a boolean indicating if a descendant already has requested name.
        """
        return (
            False
            if self.parent is None and self.name != name
            else self.parent.name_already_in_tree(name)
        )

    def to_dict(self) -> dict:
        """
        Convert self to dict.
        :param compile: if True, all models in tree nodes will be compiled. ImpactModel
        will be bigger, but its execution will be faster at first use.
        :return: self as a dict
        """
        return {
            "name": self.name,
            "models": {
                str(method): str(model) for method, model in self.models.items()
            },
            "direct_impacts": {
                str(method): str(model) for method, model in self.direct_impacts.items()
            },
            "scaled_direct_impacts": {
                str(method): str(model)
                for method, model in self.scaled_direct_impacts.items()
            },
            "children": [child.to_dict() for child in self.children],
            "properties": self.properties.properties,
            "amount": str(self.amount),
        }

    @staticmethod
    def from_dict(impact_model_tree_node: dict) -> ImpactTreeNode:
        """
        Convert dict to ImpactTreeNode object.
        :param impact_model_tree_node: dict containing construction parameters of the
        node.
        :return: constructed node
        """
        node = ImpactTreeNode(
            name=impact_model_tree_node["name"],
            models={
                method: parse_expr(expr)
                for method, expr in impact_model_tree_node["models"].items()
            },
            direct_impacts={
                method: parse_expr(expr)
                for method, expr in impact_model_tree_node["direct_impacts"].items()
            },
            scaled_direct_impacts={
                method: parse_expr(expr)
                for method, expr in impact_model_tree_node[
                    "scaled_direct_impacts"
                ].items()
            },
            properties=NodeProperties.from_dict(impact_model_tree_node["properties"]),
            amount=parse_expr(impact_model_tree_node["amount"]),
        )
        for child in impact_model_tree_node["children"]:
            node.new_child_from_dict(child)
        return node

    @staticmethod
    def node_name_to_symbol_name(node_name: str) -> str:
        """
        Convert node name to symbol name by replacing everything which is not an
        alphanumerical character by an underscore.
        :param node_name: node's name
        :return: symbol's name
        """
        return re.sub("[^0-9a-zA-Z]+", "_", node_name)

    def compute(
        self,
        transformed_params: Dict[
            str, Union[List[Union[str, float]], Union[str, float]]
        ],
        direct_impacts: bool = False,
    ) -> LCIAScores:
        """
        Compute node's impacts with given parameters values.
        Multithreading is used to compute different impact methods in parallel.
        :param transformed_params: parameters, transformed by ImpactModelParam's
        transform method.
        :param direct_impacts: if True, direct_impacts will be computed instead of
        full impacts (i.e. sum of direct impacts and children direct impacts)
        :return: a dict mapping impact's name with corresponding score, or list of
        scores.
        """
        if not self.models_compiled:
            self.compile_models()
        if direct_impacts:
            lambda_models = {
                method: lambdify(
                    [param for param in transformed_params], model, modules=["numpy"]
                )
                for method, model in self.scaled_direct_impacts.items()
            }
        else:
            lambda_models = {
                method: lambdify(
                    [param for param in transformed_params], model, modules=["numpy"]
                )
                for method, model in self.models.items()
            }
        results = {}
        with ThreadPoolExecutor() as executor:
            futures = []
            for method_name, lambda_model in lambda_models.items():
                futures.append(
                    executor.submit(
                        self._multithread_compute_process,
                        method_name=method_name,
                        lambda_model=lambda_model,
                        **transformed_params,
                    )
                )
            for future in as_completed(futures):
                results.update(future.result())

        if isinstance(list(results.values())[0], np.ndarray):
            results_ = results.copy()
            results.update({key: value.tolist() for key, value in results_.items()})
        return LCIAScores(scores=results)

    @staticmethod
    def _multithread_compute_process(method_name, lambda_model, **params):
        """
        Execute a lambda model with given params.
        :param method_name: impact method's name.
        :param lambda_model: model to execute.
        :param params: parameters of the lambda model.
        :return: a dict mapping results with method's name.
        """
        result = lambda_model(**params)
        # Lambdified model returns a constant if model is a constant itself.
        # Need to convert it to ndarray.

        if len(params)==0:
            return {method_name: result}

        if isinstance(list(params.values())[0], np.ndarray) and not isinstance(
            result, np.ndarray
        ):
            result = np.repeat(result, len(list(params.values())[0]))
        return {method_name: result}



import os
from typing import Optional

import typer
import yaml

import apparun.core

cli_app = typer.Typer()


@cli_app.command()
def compute(
    impact_model_name: str,
    params_file_path: str,
    output_file_path: Optional[str] = None,
):
    with open(params_file_path, "r") as stream:
        params = yaml.safe_load(stream) or {}
    scores = apparun.core.compute_impacts(impact_model_name, params)
    print(scores)
    if output_file_path is not None:
        with open(output_file_path, "w") as stream:
            yaml.dump(scores, stream, sort_keys=False)


@cli_app.command()
def compute_nodes(
    impact_model_name: str,
    params_file_path: str,
    output_file_path: Optional[str] = None,
):
    with open(params_file_path, "r") as stream:
        params = yaml.safe_load(stream) or {}
    scores = apparun.core.compute_impacts(impact_model_name, params, all_nodes=True)
    print(scores)
    if output_file_path is not None:
        with open(output_file_path, "w") as stream:
            yaml.dump(scores, stream, sort_keys=False)


@cli_app.command()
def models():
    valid_impact_models = apparun.core.get_valid_models()
    print(valid_impact_models)


@cli_app.command()
def model_params(impact_model_name: str):
    impact_model_params = apparun.core.get_model_params(impact_model_name)
    print(impact_model_params)


@cli_app.command()
def results(results_config_file_path: str):
    with open(results_config_file_path, "r") as stream:
        results_config = yaml.safe_load(stream)
    apparun.core.compute_results(results_config)


if __name__ == "__main__":
    cli_app()



from typing import Dict, List, Union

from fastapi import FastAPI
from pydantic import BaseModel

import apparun.core

app = FastAPI()


class ComputeParams(BaseModel):
    impact_model_name: str
    params: Dict[str, Union[str, float, List[Union[str, float]]]]


class GetModelParams(BaseModel):
    impact_model_name: str


@app.post("/compute/")
def compute(params: ComputeParams):
    scores = apparun.core.compute_impacts(params.impact_model_name, params.params)
    return scores


@app.post("/compute_nodes/")
def compute_nodes(params: ComputeParams):
    scores = apparun.core.compute_impacts(
        params.impact_model_name, params.params, all_nodes=True
    )
    return scores


@app.get("/get_models/")
def get_models():
    valid_impact_models = apparun.core.get_valid_models()
    return valid_impact_models


@app.post("/get_model_params/")
def get_model_params(params: GetModelParams):
    impact_models_params = apparun.core.get_model_params(params.impact_model_name)
    return impact_models_params



import os

import pytest

from apparun.impact_methods import MethodFullName, MethodShortName
from apparun.impact_model import ImpactModel, ModelMetadata
from apparun.impact_tree import ImpactTreeNode
from apparun.parameters import ImpactModelParams
from apparun.results import get_result


def test_impact_tree_calculation():
    impact_model = ImpactModel().from_yaml("data/noparam_system.yaml")
    lcia_scores = impact_model.get_scores()
    scores = sorted(lcia_scores.scores.items())
    assert dict(scores) == {'EFV3_CLIMATE_CHANGE': 1.3830513221576728e-06, 'EFV3_ECOTOXICITY_FRESHWATER': 4.365302017373894e-12, 'EFV3_LAND_USE': 0.0}

def test_impact_calculation_result():
    impact_model = ImpactModel().from_yaml("data/noparam_system.yaml")
    lcia_nodes_scores = impact_model.get_nodes_scores()

    print(lcia_nodes_scores)

    tree_result = get_result("tree_map")(
        impact_model=impact_model,
        n=4096,
        output_name="treemap",
        pdf_save_path=os.path.join("outputs", "figures/"),
        table_save_path=os.path.join("outputs", "tables/"),
        html_save_path=os.path.join("outputs", "figures/"),
    )

    result_table = tree_result.get_table()
    tree_result.get_figure(result_table)
    assert os.path.exists("outputs/tables/treemap.csv")
    assert os.path.exists("outputs/figures/treemap-EFV3_CLIMATE_CHANGE.pdf")
    assert os.path.exists("outputs/figures/treemap-EFV3_CLIMATE_CHANGE.html")



from __future__ import annotations

from aenum import Enum, NoAlias


class MethodFullName(str, Enum):
    """
    Enumeration of impact methods supported by Brightway.
    So far, only some ReCiPe methods, and all EF v3.0 impact methods are supported.
    """

    # EFV3
    EFV3_ACIDIFICATION = "('EF v3.0', 'acidification', 'accumulated exceedance (AE)')"
    EFV3_CLIMATE_CHANGE = (
        "('EF v3.0', 'climate change', 'global warming potential (GWP100)')"
    )
    EFV3_CLIMATE_CHANGE_BIOGENIC = (
        "('EF v3.0', 'climate change: biogenic', 'global warming potential (GWP100)')"
    )
    EFV3_CLIMATE_CHANGE_FOSSIL = (
        "('EF v3.0', 'climate change: fossil', 'global warming potential (GWP100)')"
    )
    EFV3_CLIMATE_CHANGE_LAND_USE = "('EF v3.0', 'climate change: land use and land use change', 'global warming potential (GWP100)')"
    EFV3_ECOTOXICITY_FRESHWATER = "('EF v3.0', 'ecotoxicity: freshwater', 'comparative toxic unit for ecosystems (CTUe)')"
    EFV3_ECOTOXICITY_FRESHWATER_INORGANICS = "('EF v3.0', 'ecotoxicity: freshwater, inorganics', 'comparative toxic unit for ecosystems (CTUe)')"
    EFV3_ECOTOXICITY_FRESHWATER_METALS = "('EF v3.0', 'ecotoxicity: freshwater, metals', 'comparative toxic unit for ecosystems (CTUe)')"
    EFV3_ECOTOXICITY_FRESHWATER_ORGANICS = "('EF v3.0', 'ecotoxicity: freshwater, organics', 'comparative toxic unit for ecosystems (CTUe)')"
    EFV3_EUTROPHICATION_FRESHWATER = "('EF v3.0', 'eutrophication: freshwater', 'fraction of nutrients reaching freshwater end compartment (P)')"
    EFV3_EUTROPHICATION_MARINE = "('EF v3.0', 'eutrophication: marine', 'fraction of nutrients reaching marine end compartment (N)')"
    EFV3_EUTROPHICATION_TERRESTRIAL = (
        "('EF v3.0', 'eutrophication: terrestrial', 'accumulated exceedance (AE)')"
    )
    EFV3_HUMAN_TOXICITY_CARCINOGENIC = "('EF v3.0', 'human toxicity: carcinogenic', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_CARCINOGENIC_INORGANICS = "('EF v3.0', 'human toxicity: carcinogenic, inorganics', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_CARCINOGENIC_METALS = "('EF v3.0', 'human toxicity: carcinogenic, metals', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_CARCINOGENIC_ORGANICS = "('EF v3.0', 'human toxicity: carcinogenic, organics', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC = "('EF v3.0', 'human toxicity: non-carcinogenic', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC_INORGANICS = "('EF v3.0', 'human toxicity: non-carcinogenic, inorganics', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC_METALS = "('EF v3.0', 'human toxicity: non-carcinogenic, metals', 'comparative toxic unit for human (CTUh)')"
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC_ORGANICS = "('EF v3.0', 'human toxicity: non-carcinogenic, organics', 'comparative toxic unit for human (CTUh)')"
    EFV3_IONISING_RADIATION = "('EF v3.0', 'ionising radiation: human health', 'human exposure efficiency relative to u235')"
    EFV3_LAND_USE = "('EF v3.0', 'land use', 'soil quality index')"
    EFV3_MATERIAL_RESOURCES = "('EF v3.0', 'material resources: metals/minerals', 'abiotic depletion potential (ADP): elements (ultimate reserves)')"
    EFV3_ENERGY_RESOURCES = "('EF v3.0', 'energy resources: non-renewable', 'abiotic depletion potential (ADP): fossil fuels')"
    EFV3_OZONE_DEPLETION = (
        "('EF v3.0', 'ozone depletion', 'ozone depletion potential (ODP)')"
    )
    EFV3_PARTICULATE_MATTER_FORMATION = (
        "('EF v3.0', 'particulate matter formation', 'impact on human health')"
    )
    EFV3_PHOTOCHEMICAL_OZONE_FORMATION = "('EF v3.0', 'photochemical oxidant formation: human health', 'tropospheric ozone concentration increase')"
    EFV3_WATER_USE = "('EF v3.0', 'water use', 'user deprivation potential (deprivation-weighted water consumption)')"

    def to_short_name(self) -> MethodShortName:
        """
        Convert a full impact name (as specified in Brightway) to its shorter version.
        :return: short name version of full impact name.
        """
        return MethodShortName[self.name]


class MethodShortName(str, Enum):
    """
    Short version of impact methods supported by Brightway, to ease readability of
    figures.
    So far, only some ReCiPe methods, and all EF v3.0 impact methods are supported.
    """

    _settings_ = NoAlias
    # RECIPE
    RECIPE_MIDPOINT_CLIMATE_CHANGE = "Climate change (GWP100)"
    RECIPE_MIDPOINT_TERRESTRIAL_ECOTOXICITY = "Terrestrial ecotoxicity (TETPinf)"
    RECIPE_ENDPOINT_ECOSYSTEM_QUALITY = "Damage to ecosystems"
    RECIPE_ENDPOINT_RESOURCES = "Damage to resource availability"
    RECIPE_ENDPOINT_HUMAN_HEALTH = "Damage to human health"
    RECIPE_ENDPOINT_TOTAL = "Total damage"

    # EFV3
    EFV3_ACIDIFICATION = "Acification (AE)"
    EFV3_CLIMATE_CHANGE = "Climate change (GWP100)"
    EFV3_CLIMATE_CHANGE_BIOGENIC = "Climate change: biogenic (GWP100)"
    EFV3_CLIMATE_CHANGE_FOSSIL = "Climate change: fossil (GWP100)"
    EFV3_CLIMATE_CHANGE_LAND_USE = "Climate change: land use/land use change (GWP100)"
    EFV3_ECOTOXICITY_FRESHWATER = "Ecotoxicity: freshwater (CTUe)"
    EFV3_ECOTOXICITY_FRESHWATER_INORGANICS = (
        "Ecotoxicity: freshwater, inorganics (CTUe)"
    )
    EFV3_ECOTOXICITY_FRESHWATER_METALS = "Ecotoxicity: freshwater, metals (CTUe)"
    EFV3_ECOTOXICITY_FRESHWATER_ORGANICS = "Ecotoxicity: freshwater, organics (CTUe)"
    EFV3_EUTROPHICATION_FRESHWATER = "Eutrophication: freshwater (kgPeq)"
    EFV3_EUTROPHICATION_MARINE = "Eutrophication: marine (N)"
    EFV3_EUTROPHICATION_TERRESTRIAL = "Eutrophication: terrestrial (AE)"
    EFV3_HUMAN_TOXICITY_CARCINOGENIC = "Human toxicity: carcinogenic (CTUh)"
    EFV3_HUMAN_TOXICITY_CARCINOGENIC_INORGANICS = (
        "Human toxicity: carcinogenic, inorganics (CTUh)"
    )
    EFV3_HUMAN_TOXICITY_CARCINOGENIC_METALS = (
        "Human toxicity: carcinogenic, metals (CTUh)"
    )
    EFV3_HUMAN_TOXICITY_CARCINOGENIC_ORGANICS = (
        "Human toxicity: carcinogenic, organics (CTUh)"
    )
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC = "Human toxicity: non-carcinogenic (CTUh)"
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC_INORGANICS = (
        "Human toxicity: non-carcinogenic, inorganics (CTUh)"
    )
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC_METALS = (
        "Human toxicity: non-carcinogenic, metals (CTUh)"
    )
    EFV3_HUMAN_TOXICITY_NON_CARCINOGENIC_ORGANICS = (
        "Human toxicity: non-carcinogenic, organics (CTUh)"
    )
    EFV3_IONISING_RADIATION = "Ionising radiation: human health (kBqU235)"
    EFV3_LAND_USE = "Land use (soil quality index)"
    EFV3_MATERIAL_RESOURCES = "Resource use, metals and minerals (kgSbeq)"
    EFV3_ENERGY_RESOURCES = "Energy use, energy carriers (MJ)"
    EFV3_OZONE_DEPLETION = "Ozone depletion (ODP)"
    EFV3_PARTICULATE_MATTER_FORMATION = (
        "Particulate matter formation: impact on human health (disease incidences)"
    )
    EFV3_PHOTOCHEMICAL_OZONE_FORMATION = "Photochemical ozone formation (kgNMVOCeq)"
    EFV3_WATER_USE = "Depr.-weighted water cons. (kg world eq. deprived)"

    def to_full_name(self) -> MethodFullName:
        """
        Convert a short impact name to its full version (as specified in Brightway).
        :return: full name version of short impact name.
        """
        return MethodFullName[self.name]



from __future__ import annotations

import re
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from pydantic import BaseModel
from SALib.sample import sobol


class ImpactModelParam(BaseModel):
    """
    Impact model parameter.
    """

    name: str
    default: Union[str, float]

    def to_dict(self) -> dict:
        """
        Convert self to dict.
        :return: self as a dict
        """
        return dict(self)

    def name_is_valid(self) -> bool:
        """
        Check if parameter's name is considered as valid.
        Parameter name should only be composed of letters, numbers, underscores, and
        contain at least one letter.
        :return: if parameter's name is valid
        """
        return bool(re.match("^[A-Za-z0-9_]*$", self.name)) and bool(
            re.match("[A-Za-z]+", self.name)
        )

    @staticmethod
    def from_dict(impact_model_param: dict) -> ImpactModelParam:
        """
        Convert dict to ImpactModelParam object. Subclass must be specified in 'type'
        field.
        :param impact_model_param: dict containing construction parameters of the param.
        :param frozen: should base model be frozen? Allows it to be hashable.
        :return: constructed param
        """
        if impact_model_param["type"] == "float":
            param = FloatParam(**impact_model_param)
        elif impact_model_param["type"] == "enum":
            param = EnumParam(**impact_model_param)
        else:
            raise ValueError(f"Unknown parameter type {impact_model_param['type']}")
        if not param.name_is_valid():
            raise ValueError(
                f"Parameter name {param.name} is not valid. Check ImpactModelParam."
                f"name_is_valid docstring to know which name is considered valid/"
            )
        return param

    def update_default(self, new_value: Union[float, str]):
        """
        Change default value of the parameter, and compute new bounds if relevant.
        :param new_value: new default value
        """
        self.default = new_value

    def draw_to_distrib(self, samples) -> Union[List[str], List[float]]:
        return

    def corresponds(self, symbol_name: str) -> bool:
        return


class FloatParam(ImpactModelParam):
    """
    Impact model float parameter.
    Contains a default, min and max values, as well as a distribution for monte carlo
    runs.
    Supported distributions is "linear".
    """

    type: str = "float"
    min: Optional[float] = None
    max: Optional[float] = None
    distrib: Optional[str] = "linear"
    pm: Optional[float] = None
    pm_perc: Optional[float] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.update_bounds()

    def update_default(self, new_value: float):
        """
        Change default value of the parameter, and compute new min and max values if they
        are defined using pm or pm_perc.
        :param new_value: new default value
        """
        super().update_default(new_value)
        self.update_bounds()

    def update_bounds(self):
        """
        Compute min and max attributes from default attribute if pm, or pm_perc attributes
        is not None.
        """
        if self.pm is not None:
            self.min = self.default - self.pm
            self.max = self.default + self.pm
        if self.pm_perc is not None:
            self.min = self.default - (self.default * self.pm_perc)
            self.max = self.default + (self.default * self.pm_perc)

    def transform(
        self, values: Union[float, List[float]]
    ) -> Dict[str, Union[float, np.array]]:
        """
        Transform float value, or list of float values to be readily usable by
        ImpactModel.
        Value(s) are mapped to a dict with parameter name as a key.
        :param values:
        :return: a dict mapping parameter name and transformed values.
        """
        if isinstance(values, (int, float)):
            return {self.name: values}
        return {self.name: np.array(values)}

    def draw_to_distrib(self, samples: np.ndarray) -> List[float]:
        if self.distrib == "linear":
            return list(self.min + samples * (self.max - self.min))

    def corresponds(self, symbol_name: str) -> bool:
        return symbol_name == self.name


class EnumParam(ImpactModelParam):
    """
    Impact model enum parameter.
    Contains a default, list of options and associated weights for monte carlo run.
    runs.
    """

    type: str = "enum"
    weights: Dict[str, float]

    @property
    def dummies_names(self):
        """
        Give dummies names of all possible options.
        A dummy name is the option with parameter name as a prefix.
        :return: a list containing dummies names for all possible options.
        """
        return [self.full_option_name(option) for option in self.options]

    @property
    def options(self):
        return self.weights.keys()

    def full_option_name(self, option: str) -> str:
        """
        Return option with parameter name as a prefix.
        :param option: enum option.
        :return: option name with parameter name as a prefix.
        """
        return f"{self.name}_{option}"

    def look_up_table(self) -> Dict[str, Dict[str, int]]:
        """
        Map every option to its one hot encoded vector, i.e. a dict with dummies names
        and associated value (1 if corresponding to current option, 0 otherwise).
        :return: a nested dict with all option and corresponding one hot encoded vector.
        """
        return {
            option: {
                self.full_option_name(possible_option): int(option == possible_option)
                for possible_option in self.options
            }
            for option in self.options
        }

    def transform(
        self, values: Union[str, List[str]]
    ) -> Dict[str, Union[float, np.array]]:
        """
        Transform option, or list of options to be readily usable by ImpactModel.
        Value(s) one hot encoded, then mapped to a dict with parameter name as a key.
        :param values:
        :return: a dict mapping parameter name and transformed values.
        """
        if isinstance(values, list):
            return (
                pd.DataFrame([self.look_up_table().get(value) for value in values])
                .apply(np.array, axis=0, result_type="reduce")
                .to_dict()
            )
        return self.look_up_table()[values]

    def draw_to_distrib(self, samples: np.ndarray) -> List[str]:
        bin_to_names = {
            i + 1: list(self.weights.keys())[i] for i in range(len(self.weights))
        }
        weights = list(self.weights.values())
        bins = [0] + list(np.cumsum(weights) / np.sum(weights))
        transformed_samples = list(np.digitize(samples, bins))
        transformed_samples = list(
            pd.DataFrame(transformed_samples).replace(bin_to_names).values.ravel()
        )
        return transformed_samples

    def corresponds(self, symbol_name: str) -> bool:
        return symbol_name in self.dummies_names


class ImpactModelParams(BaseModel):
    parameters: List[ImpactModelParam]
    sobol_problem: Optional[Dict] = None

    @property
    def names(self):
        return [parameter.name for parameter in self.parameters]

    def to_list(self, sorted_by_name: Optional[bool] = False) -> List[Dict]:
        """
        Convert each parameter of parameters attribute to a dict, return them as a list,
        optionally sorted by parameter's name
        :param sorted_by_name: sort parameters list by parameter's name if True
        :return: list of parameter, themselves as dict
        """
        if sorted_by_name:
            sorted_names = sorted([parameter.name for parameter in self.parameters])
            return [self.get_parameter_by_name(name).to_dict() for name in sorted_names]
        return [param.to_dict() for param in self.parameters]

    def update_defaults(self, defaults: Dict[str, Union[str, float]]):
        """
        Change default value of the parameters contained in argument, and compute new
        bounds if relevant.
        :param defaults: dict containing parameter name as key and new default as
        value
        """
        for parameter in self.parameters:
            if parameter.name in defaults.keys():
                parameter.update_default(defaults[parameter.name])

    def get_parameter_by_name(self, parameter_name: str) -> Optional[ImpactModelParam]:
        matching_param = [
            parameter for parameter in self if parameter.name == parameter_name
        ]
        if len(matching_param) == 0:
            return None
        return matching_param[0]

    def get_missing_parameter_names(self, parameters: Union[List, Dict]) -> List:
        known_parameters = (
            list(parameters) if isinstance(parameters, dict) else parameters
        )
        return [
            parameter for parameter in self.names if parameter not in known_parameters
        ]

    def get_default_values(self, parameter_names: List) -> Dict:
        parameters = [
            self.get_parameter_by_name(parameter_name)
            for parameter_name in parameter_names
        ]
        parameters = {parameter.name: parameter.default for parameter in parameters}
        return parameters

    def find_corresponding_parameter(
        self, symbol_name: str, must_find_one: Optional[bool] = True
    ) -> Union[ImpactModelParam, List[ImpactModelParam]]:
        matching_parameters = [
            parameter
            for parameter in self.parameters
            if parameter.corresponds(symbol_name)
        ]
        if must_find_one:
            if len(matching_parameters) > 1:
                raise ValueError(
                    f"{symbol_name} matches with multiple params ({[matching_parameter.name for matching_parameter in matching_parameters]})."
                )
            if len(matching_parameters) < 1:
                raise ValueError(f"{symbol_name} doesn't match with any params.")
            return matching_parameters[0]
        if len(matching_parameters) == 0:
            return []
        return matching_parameters

    @staticmethod
    def from_list(parameters) -> ImpactModelParams:
        return ImpactModelParams(
            parameters=[
                param
                if isinstance(param, ImpactModelParam)
                else ImpactModelParam.from_dict(param)
                for param in parameters
            ]
        )

    def __init__(self, **args):
        super().__init__(**args)
        self.__curr = 0

    def __iter__(self):
        return self

    def __next__(self) -> ImpactModelParam:
        if self.__curr >= len(self.parameters):
            self.__curr = 0
            raise StopIteration()

        param = self.parameters[self.__curr]
        self.__curr += 1
        return param

    def __getitem__(self, i) -> ImpactModelParam:
        return self.parameters[i]

    def set_sobol_problem(self):
        self.sobol_problem = {
            "num_vars": len(self.parameters),
            "names": [parameter.name for parameter in self.parameters],
            "bounds": [[0, 1]] * len(self.parameters),
        }

    def sobol_draw(self, n) -> np.ndarray:
        self.set_sobol_problem()
        samples = sobol.sample(self.sobol_problem, n)
        return samples

    def uniform_draw(self, n) -> np.ndarray:
        return np.random.rand(n, len(self.parameters))

    def draw_to_distrib(
        self, samples: np.ndarray
    ) -> Dict[str, Union[List[float], List[str]]]:
        return {
            self.parameters[i].name: self.parameters[i].draw_to_distrib(samples[:, i])
            for i in range(len(self.parameters))
        }
